__version__ = '0.4.0'
__author__ = 'Oleg Kostromin, Marco Pasini, Iryna Kondrashchenko'

from falcon.main import initialize, AutoML
from falcon.utils import disable_warnings, run_model

disable_warnings()